@@ -663,7 +663,7 @@ cram_block *cram_encode_slice_header(cram_fd *fd, cram_slice *s) {
     if (!b)
 	return NULL;
 
-    if (NULL == (cp = buf = malloc(16+5*(8+s->hdr->num_blocks)))) {
+    if (NULL == (cp = buf = malloc(22+16+5*(8+s->hdr->num_blocks)))) {
 	cram_free_block(b);
 	return NULL;
     }
@@ -686,7 +686,25 @@ cram_block *cram_encode_slice_header(cram_fd *fd, cram_slice *s) {
 	memcpy(cp, s->hdr->md5, 16); cp += 16;
     }
     
-    assert(cp-buf <= 16+5*(8+s->hdr->num_blocks));
+    if (CRAM_MAJOR_VERS(fd->version) >= 3) {
+	if (s->BD_crc || s->SD_crc) {
+	    *cp++ = 'B'; *cp++ = 'D'; *cp++ = 'B'; *cp++ = 'c';
+	    *cp++ = 4; *cp++ = 0; *cp++ = 0; *cp++ = 0;
+	    *cp++ = (s->BD_crc >>  0) & 0xff;
+	    *cp++ = (s->BD_crc >>  8) & 0xff;
+	    *cp++ = (s->BD_crc >> 16) & 0xff;
+	    *cp++ = (s->BD_crc >> 24) & 0xff;
+
+	    *cp++ = 'S'; *cp++ = 'D'; *cp++ = 'B'; *cp++ = 'c';
+	    *cp++ = 4; *cp++ = 0; *cp++ = 0; *cp++ = 0;
+	    *cp++ = (s->SD_crc >>  0) & 0xff;
+	    *cp++ = (s->SD_crc >>  8) & 0xff;
+	    *cp++ = (s->SD_crc >> 16) & 0xff;
+	    *cp++ = (s->SD_crc >> 24) & 0xff;
+	}
+    }
+
+    assert(cp-buf <= 22+16+5*(8+s->hdr->num_blocks));
 
     b->data = (unsigned char *)buf;
     b->comp_size = b->uncomp_size = cp-buf;
@@ -856,7 +874,8 @@ static int cram_encode_slice_read(cram_fd *fd,
 	    case 'b':
 		// string of bases
 		r |= h->codecs[DS_BB]->encode(s, h->codecs[DS_BB], 
-					      BLOCK_DATA(s->seqs_blk) + f->b.seq_idx,
+					      (char *)BLOCK_DATA(s->seqs_blk)
+                                                      + f->b.seq_idx,
 					      f->b.len);
 		break;
 
@@ -897,7 +916,8 @@ static int cram_encode_slice_read(cram_fd *fd,
 				      (char *)&cr->mqual, 1);
     } else {
 	char *seq = (char *)BLOCK_DATA(s->seqs_blk) + cr->seq;
-	r |= h->codecs[DS_BA]->encode(s, h->codecs[DS_BA], seq, cr->len);
+	if (cr->len)
+	    r |= h->codecs[DS_BA]->encode(s, h->codecs[DS_BA], seq, cr->len);
     }
 
     return r ? -1 : 0;
@@ -1125,6 +1145,8 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
 		}
 		break;
 	    }
+	    default:
+		break;
 	    }
 	} else {
 	    if (!(id == DS_BB && !h->codecs[DS_BB]))
@@ -1212,9 +1234,9 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
     spare_bams *spares;
 
     /* Cache references up-front if we have unsorted access patterns */
-    pthread_mutex_lock(&fd->ref_lock);
+    if (fd->ref_lock) pthread_mutex_lock(fd->ref_lock);
     nref = fd->refs->nref;
-    pthread_mutex_unlock(&fd->ref_lock);
+    if (fd->ref_lock) pthread_mutex_unlock(fd->ref_lock);
 
     if (!fd->no_ref && c->refs_used) {
 	for (i = 0; i < nref; i++) {
@@ -1277,14 +1299,16 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 		    }
 
 		    c->ref_seq_id = bam_ref(b); // overwritten later by -2
-		    assert(fd->refs->ref_id[c->ref_seq_id]->seq);
+		    if (!fd->refs->ref_id[c->ref_seq_id]->seq)
+			return -1;
 		    c->ref       = fd->refs->ref_id[c->ref_seq_id]->seq;
 		    c->ref_start = 1;
 		    c->ref_end   = fd->refs->ref_id[c->ref_seq_id]->length;
 		}
 	    }
 
-	    process_one_read(fd, c, s, cr, b, r2);
+	    if (process_one_read(fd, c, s, cr, b, r2) != 0)
+		return -1;
 
 	    if (first_base > cr->apos)
 		first_base = cr->apos;
@@ -1312,11 +1336,11 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 
     /* Link our bams[] array onto the spare bam list for reuse */
     spares = malloc(sizeof(*spares));
-    pthread_mutex_lock(&fd->bam_list_lock);
+    if (fd->bam_list_lock) pthread_mutex_lock(fd->bam_list_lock);
     spares->bams = c->bams;
     spares->next = fd->bl;
     fd->bl = spares;
-    pthread_mutex_unlock(&fd->bam_list_lock);
+    if (fd->bam_list_lock) pthread_mutex_unlock(fd->bam_list_lock);
     c->bams = NULL;
 
     /* Detect if a multi-seq container */
@@ -1352,7 +1376,7 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
     }
 
     c->num_records = 0;
-    c->num_blocks = 0;
+    c->num_blocks = 1; // cram_block_compression_hdr
     c->length = 0;
 
     if (fd->verbose > 1) fprintf(stderr, "BF_stats: ");
@@ -1609,7 +1633,7 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
     for (i = 0; i < c->curr_slice; i++) {
 	cram_slice *s = c->slices[i];
 	
-	c->num_blocks += s->hdr->num_blocks + 2;
+	c->num_blocks += s->hdr->num_blocks + 1; // slice header
 	c->landmark[i] = slice_offset;
 
 	if (s->hdr->ref_seq_start + s->hdr->ref_seq_span >
@@ -1720,7 +1744,6 @@ static int cram_add_bases(cram_fd *fd, cram_container *c,
 			  cram_slice *s, cram_record *r,
 			  int pos, int len, char *base) {
     cram_feature f;
-    int i;
 
     f.b.pos = pos+1;
     f.b.code = 'b';
@@ -2029,25 +2052,34 @@ static char *cram_encode_aux(cram_fd *fd, bam_seq_t *b, cram_container *c,
     while (aux[0] != 0 && aux - orig < aux_size) {
 	HashData hd; hd.i = 0;
 
+	// RG:Z
 	if (aux[0] == 'R' && aux[1] == 'G' && aux[2] == 'Z') {
 	    rg = &aux[3];
 	    while (*aux++);
 	    continue;
 	}
+
+	// MD:Z
 	if (aux[0] == 'M' && aux[1] == 'D' && aux[2] == 'Z') {
-	    while (*aux++);
-	    continue;
+	    if (cr->len && !fd->no_ref && !(cr->flags & BAM_FUNMAP)) {
+		while (*aux++);
+		continue;
+	    }
 	}
+
+	// NM:i
 	if (aux[0] == 'N' && aux[1] == 'M') {
-	    switch(aux[2]) {
-	    case 'A': case 'C': case 'c': aux+=4; break;
-	    case 'S': case 's':           aux+=5; break;
-	    case 'I': case 'i': case 'f': aux+=7; break;
-	    default:
-		fprintf(stderr, "Unhandled type code for NM tag\n");
-		return NULL;
+	    if (cr->len && !fd->no_ref && !(cr->flags & BAM_FUNMAP)) {
+		switch(aux[2]) {
+		case 'A': case 'C': case 'c': aux+=4; break;
+		case 'S': case 's':           aux+=5; break;
+		case 'I': case 'i': case 'f': aux+=7; break;
+		default:
+		    fprintf(stderr, "Unhandled type code for NM tag\n");
+		    return NULL;
+		}
+		continue;
 	    }
-	    continue;
 	}
 
 	BLOCK_APPEND(td_b, aux, 3);
@@ -2428,7 +2460,8 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     // FIXME: multi-ref containers
 
     ref = c->ref;
-    cr->len         = bam_seq_len(b); cram_stats_add(c->stats[DS_RL], cr->len);
+    cr->flags       = bam_flag(b);
+    cr->len         = bam_seq_len(b);
 
     //fprintf(stderr, "%s => %d\n", rg ? rg : "\"\"", cr->rg);
 
@@ -2452,7 +2485,8 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	cr->rg = brg ? brg->id : -1;
     } else if (IS_CRAM_1_VERS(fd)) {
 	SAM_RG *brg = sam_hdr_find_rg(fd->header, "UNKNOWN");
-	assert(brg);
+	if (!brg)
+	    return -1;
     } else {
 	cr->rg = -1;
     }
@@ -2460,9 +2494,6 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 
     
     cr->ref_id      = bam_ref(b);  cram_stats_add(c->stats[DS_RI], cr->ref_id);
-    cr->flags       = bam_flag(b);
-    if (bam_cigar_len(b) == 0)
-	cr->flags |= BAM_FUNMAP;
     cram_stats_add(c->stats[DS_BF], fd->cram_flag_swap[cr->flags & 0xfff]);
 
     // Non reference based encoding means storing the bases verbatim as features, which in
@@ -2471,6 +2502,9 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	cr->cram_flags = CRAM_FLAG_PRESERVE_QUAL_SCORES;
     else
 	cr->cram_flags = 0;
+
+    if (cr->len <= 0 && CRAM_MAJOR_VERS(fd->version) >= 3)
+	cr->cram_flags |= CRAM_FLAG_NO_SEQ;
     //cram_stats_add(c->stats[DS_CF], cr->cram_flags);
 
     c->num_bases   += cr->len;
@@ -2576,6 +2610,9 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 
     qual = cp = (char *)bam_qual(b);
 
+    if (CRAM_MAJOR_VERS(fd->version) >= 3 && !fd->ignore_chksum)
+	s->BD_crc += crc32(0L, (Bytef *) seq,  cr->len);
+
     /* Copy and parse */
     if (!(cr->flags & BAM_FUNMAP)) {
 	int32_t *cig_to, *cig_from;
@@ -2618,6 +2655,11 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 		    char *sp = &seq[spos];
 		    char *rp = &ref[apos];
 		    char *qp = &qual[spos];
+		    if (end > cr->len) {
+			fprintf(stderr, "CIGAR and query sequence are of "
+				"different length\n");
+			return -1;
+		    }
 		    for (l = 0; l < end; l++) {
 			if (rp[l] != sp[l]) {
 			    if (!sp[l])
@@ -2738,8 +2780,17 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 		if (cram_add_pad(c, s, cr, spos, cig_len, &seq[spos]))
 		    return -1;
 		break;
+
+	    default:
+		fprintf(stderr, "Unknown CIGAR op code %d\n", cig_op);
+		return -1;
 	    }
 	}
+	if (cr->len && spos != cr->len) {
+	    fprintf(stderr, "CIGAR and query sequence are of different "
+		    "length\n");
+	    return -1;
+	}
 	fake_qual = spos;
 	cr->aend = MIN(apos, c->ref_end);
 	cram_stats_add(c->stats[DS_FN], cr->nfeature);
@@ -2752,6 +2803,7 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	cr->aend = cr->apos;
 	for (i = 0; i < cr->len; i++)
 	    cram_stats_add(c->stats[DS_BA], seq[i]);
+	fake_qual = 0;
     }
 
     /*
@@ -2762,7 +2814,7 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     if (cr->cram_flags & CRAM_FLAG_PRESERVE_QUAL_SCORES) {
 	/* Special case of seq "*" */
 	if (cr->len == 0) {
-	    cram_stats_add(c->stats[DS_RL], cr->len = fake_qual);
+	    cr->len = fake_qual;
 	    BLOCK_GROW(s->qual_blk, cr->len);
 	    cp = (char *)BLOCK_END(s->qual_blk);
 	    memset(cp, 255, cr->len);
@@ -2779,16 +2831,20 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	    } else {
 		memcpy(to, from, cr->len);
 	    }
+
+	    if (CRAM_MAJOR_VERS(fd->version) >= 3 && !fd->ignore_chksum)
+		s->SD_crc += crc32(0L, (Bytef *) to, cr->len);
+
 	    //for (i = 0; i < cr->len; i++) cp[i] = from[i];
 	}
 	BLOCK_SIZE(s->qual_blk) += cr->len;
     } else {
-	if (cr->len == 0) {
+	if (cr->len == 0)
 	    cr->len = fake_qual >= 0 ? fake_qual : cr->aend - cr->apos + 1;
-	    cram_stats_add(c->stats[DS_RL], cr->len);
-	}
     }
 
+    cram_stats_add(c->stats[DS_RL], cr->len);
+
     /* Now we know apos and aend both, update mate-pair information */
     {
 	int new;
@@ -3010,9 +3066,9 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 	    c->pos_sorted = 0; // required atm for multi_seq slices
 
 	    if (!c->refs_used) {
-		pthread_mutex_lock(&fd->ref_lock);
+		if (fd->ref_lock) pthread_mutex_lock(fd->ref_lock);
 		c->refs_used = calloc(fd->refs->nref, sizeof(int));
-		pthread_mutex_unlock(&fd->ref_lock);
+		if (fd->ref_lock) pthread_mutex_unlock(fd->ref_lock);
 		if (!c->refs_used)
 		    return -1;
 	    }
@@ -3022,20 +3078,20 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 	c->slice_rec = c->curr_rec;
 
 	// Have we seen this reference before?
-	if (bam_ref(b) >= 0 && bam_ref(b) != curr_ref && !fd->embed_ref &&
+	if (bam_ref(b) >= 0 && curr_ref >= 0 && bam_ref(b) != curr_ref && !fd->embed_ref &&
 	    !fd->unsorted && multi_seq) {
 	    
 	    if (!c->refs_used) {
-		pthread_mutex_lock(&fd->ref_lock);
+		if (fd->ref_lock) pthread_mutex_lock(fd->ref_lock);
 		c->refs_used = calloc(fd->refs->nref, sizeof(int));
-		pthread_mutex_unlock(&fd->ref_lock);
+		if (fd->ref_lock) pthread_mutex_unlock(fd->ref_lock);
 		if (!c->refs_used)
 		    return -1;
 	    } else if (c->refs_used && c->refs_used[bam_ref(b)]) {
 		fprintf(stderr, "Unsorted mode enabled\n");
-		pthread_mutex_lock(&fd->ref_lock);
+		if (fd->ref_lock) pthread_mutex_lock(fd->ref_lock);
 		fd->unsorted = 1;
-		pthread_mutex_unlock(&fd->ref_lock);
+		if (fd->ref_lock) pthread_mutex_unlock(fd->ref_lock);
 		fd->multi_seq = 1;
 	    }
 	}
@@ -3046,7 +3102,7 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 
     if (!c->bams) {
 	/* First time through, allocate a set of bam pointers */
-	pthread_mutex_lock(&fd->bam_list_lock);
+	if (fd->bam_list_lock) pthread_mutex_lock(fd->bam_list_lock);
 	if (fd->bl) {
 	    spare_bams *spare = fd->bl;
 	    c->bams = spare->bams;
@@ -3057,7 +3113,7 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 	    if (!c->bams)
 		return -1;
 	}
-	pthread_mutex_unlock(&fd->bam_list_lock);
+	if (fd->bam_list_lock) pthread_mutex_unlock(fd->bam_list_lock);
     }
 
     /* Copy or alloc+copy the bam record, for later encoding */
