@@ -61,6 +61,7 @@ static inline void RansEncInit(RansState* r)
     *r = RANS_BYTE_L;
 }
 
+#if 0 /* Curently unused */
 // Renormalize the encoder. Internal function.
 static inline RansState RansEncRenorm(RansState x, uint8_t** pptr, uint32_t freq, uint32_t scale_bits)
 {
@@ -91,6 +92,7 @@ static inline void RansEncPut(RansState* r, uint8_t** pptr, uint32_t start, uint
     // x = C(s,x)
     *r = ((x / freq) << scale_bits) + (x % freq) + start;
 }
+#endif /* Currently unused */
 
 // Flushes the rANS encoder.
 static inline void RansEncFlush(RansState* r, uint8_t** pptr)
@@ -294,6 +296,7 @@ static inline void RansDecAdvanceSymbol(RansState* r, uint8_t** pptr, RansDecSym
     RansDecAdvance(r, pptr, sym->start, sym->freq, scale_bits);
 }
 
+#if 0 /* Currently unused */
 // Advances in the bit stream by "popping" a single symbol with range start
 // "start" and frequency "freq". All frequencies are assumed to sum to "1 << scale_bits".
 // No renormalization or output happens.
@@ -311,6 +314,7 @@ static inline void RansDecAdvanceSymbolStep(RansState* r, RansDecSymbol const* s
 {
     RansDecAdvanceStep(r, sym->start, sym->freq, scale_bits);
 }
+#endif /* Currently unused */
 
 // Renormalize.
 static inline void RansDecRenorm(RansState* r, uint8_t** pptr)
@@ -572,6 +576,9 @@ unsigned char *rans_uncompress_O0(unsigned char *in, unsigned int in_size,
 	}
 	D.fc[j].C = x;
 
+	if (x+D.fc[j].F > TOTFREQ)
+	    return NULL;
+
 	RansDecSymbolInit(&syms[j], D.fc[j].C, D.fc[j].F);
 
 	/* Build reverse lookup table */
@@ -692,8 +699,7 @@ unsigned char *rans_uncompress_O0(unsigned char *in, unsigned int in_size,
 
 unsigned char *rans_compress_O1(unsigned char *in, unsigned int in_size,
 				unsigned int *out_size) {
-    unsigned char *out_buf;
-    unsigned char *cp = out_buf, *out_end;
+    unsigned char *out_buf, *out_end, *cp;
     unsigned int last_i, tab_size, rle_i, rle_j;
     RansEncSymbol syms[256][256];
 
@@ -932,6 +938,9 @@ unsigned char *rans_uncompress_O1(unsigned char *in, unsigned int in_size,
 	    if (!D[i].fc[j].F)
 		D[i].fc[j].F = TOTFREQ;
 
+	    if (x+D[i].fc[j].F > TOTFREQ)
+		return NULL;
+
 	    RansDecSymbolInit(&syms[i][j], D[i].fc[j].C, D[i].fc[j].F);
 
 	    /* Build reverse lookup table */
@@ -1060,6 +1069,10 @@ unsigned char *rans_compress(unsigned char *in, unsigned int in_size,
 
 unsigned char *rans_uncompress(unsigned char *in, unsigned int in_size,
 			       unsigned int *out_size) {
+    /* Both rans_uncompress functions need to be able to read at least 9
+       bytes. */
+    if (in_size < 9)
+        return NULL;
     return in[0]
 	? rans_uncompress_O1(in, in_size, out_size)
 	: rans_uncompress_O0(in, in_size, out_size);
