@@ -378,9 +378,14 @@ int main(int argc, char **argv) {
 	    return 1;
     }
 
-    if (ignore_md5)
+    if (ignore_md5) {
 	if (scram_set_option(in, CRAM_OPT_IGNORE_MD5, ignore_md5))
 	    return 1;
+	if (scram_set_option(in, CRAM_OPT_IGNORE_CHKSUM, ignore_md5))
+	    return 1;
+	if (scram_set_option(out, CRAM_OPT_IGNORE_CHKSUM, ignore_md5))
+	    return 1;
+    }
     
     if (sam_fields)
 	scram_set_option(in, CRAM_OPT_REQUIRED_FIELDS, sam_fields);
@@ -454,26 +459,35 @@ int main(int argc, char **argv) {
 		break;
     }
 
-    if (max_reads == -1) {
-	switch(scram_eof(in)) {
-	case 0:
+    switch(scram_eof(in)) {
+    case -1:
+	fprintf(stderr, "Failed to decode sequence\n");
+	return 1;
+    case 0:
+	if (max_reads == -1) {
 	    fprintf(stderr, "Failed to decode sequence\n");
 	    return 1;
-	case 2:
-	    fprintf(stderr, "Warning: no end-of-file block identified. "
-		    "File may be truncated.\n");
-	    break;
-	case 1: default:
-	    // expected case
+	} else {
 	    break;
 	}
+    case 2:
+	fprintf(stderr, "Warning: no end-of-file block identified. "
+		"File may be truncated.\n");
+	break;
+    case 1: default:
+	// expected case
+	break;
     }
 
     /* Finally tidy up and close files */
-    if (scram_close(in))
+    if (scram_close(in)) {
+	fprintf(stderr, "Failed in scram_close(in)\n");
 	return 1;
-    if (scram_close(out))
+    }
+    if (scram_close(out)) {
+	fprintf(stderr, "Failed in scram_close(out)\n");
 	return 1;
+    }
 
     if (p)
 	t_pool_destroy(p, 0);
